class Torlan_LinkGun extends UTWeap_LinkGun;

var float PawnDamageMult;
var float PawnFirerateMult;

var transient float LastStrengthCalc;
var float MinTimeBetweenCalcs;

// for HUD drawing
var float LinkPulseTime; 
var int LastLinkStrength;

replication {
    if (bNetDirty)
        PawnFirerateMult; // for animation
}


function FindLinkedWeapons() {
    // don't need it
    ClearTimer('FindLinkedWeapons');
}

function Torlan_LinkGun GetNextGun() {
    return (Pawn(LinkedTo) != None) ? 
        Torlan_LinkGun(Pawn(LinkedTo).Weapon) : 
        None; 
}

function Link(Torlan_LinkGun OtherGun) {
    LinkedTo = OtherGun.Instigator;
    
    OtherGun.LinkedList.AddItem(self);
    while (OtherGun != None) {
        OtherGun.CalcLinkStrengthForce();
        OtherGun = OtherGun.GetNextGun();
    }
}


function UnLink(optional bool bDelayed) {
    local Torlan_LinkGun OtherGun;

    if (!bDelayed) {
        OtherGun = GetNextGun();
        if (OtherGun != None) {
            OtherGun.LinkedList.RemoveItem(self);
            while (OtherGun != None) {
                OtherGun.CalcLinkStrengthForce();
                OtherGun = OtherGun.GetNextGun();
            }
        }

    }
    super.UnLink(bDelayed);
}

simulated state WeaponPuttingDown {
    simulated function WeaponIsDown() {
        local Torlan_LinkGun OtherGun;
        
        foreach LinkedList(OtherGun) {
            if (Pawn(OtherGun.LinkedTo) == Instigator) {
                OtherGun.Unlink();
                break;
            }
        }
        
        Super.WeaponIsDown();
    }
}

function bool AttemptLinkToPawn(UTPawn P, PrimitiveComponent HitComponent) {
    local Torlan_LinkGun OtherGun;
    local array<UTWeap_LinkGun> LinkedWeapons;
    
    if (P != None && 
        WorldInfo.Game.GameReplicationInfo.OnSameTeam(P, Instigator))
    {
        OtherGun = Torlan_LinkGun(P.Weapon);
        
        if (OtherGun != None) {
            if (LinkedTo == P) return true;
            GetLinkedWeapons(LinkedWeapons);
            if (LinkedWeapons.Find(OtherGun) == INDEX_NONE) {
                UnLink();
                Link(OtherGun);
                LinkedComponent = HitComponent;
                return true;
            }
        }

    }
    
    return false;
}


// copy-paste from Epic's code
function AttemptLinkTo(Actor Who, PrimitiveComponent HitComponent) {
    local UTVehicle UTV;
    local UTPawn UTP;
    local UTOnslaughtObjective UTO;
    local Vector         StartTrace, EndTrace, V, HitLocation, HitNormal;
    local Actor            HitActor;
    

    // redirect to vehicle if owned by a vehicle and the vehicle allows it
    if( Who != none )
    {
        UTV = UTVehicle(Who.Owner);
        if (UTV != None && UTV.AllowLinkThroughOwnedActor(Who))
        {
            Who = UTV;
        }
    }

    // Check for linking to pawns
    UTV = UTVehicle(Who);
    if (UTV != None && UTV.bValidLinkTarget)
    {
        // Check teams to make sure they are on the same side or empty
        if (
            WorldInfo.Game.GameReplicationInfo.OnSameTeam(UTV,Instigator) ||
            (!UTV.bTeamLocked && UTV.CanEnterVehicle(Instigator))
        )
        {
            if ( !WorldInfo.Game.GameReplicationInfo.OnSameTeam(UTV,Instigator)
                && (Instigator.GetTeamNum() != 255) )
            {
                UTV.SetTeamNum( Instigator.GetTeamNum() );
            }
            LinkedComponent = HitComponent;
            if ( LinkedTo != UTV )
            {
                UnLink();
                LinkedTo = UTV;
                UTV.IncrementLinkedToCount();
            }
        } else {
            // Enemy got in the way, break any links
            UnLink();
        }
    }
    else
    {
        UTO = UTOnslaughtObjective(Who);
        if ((UTO != none) && WorldInfo.Game.GameReplicationInfo.OnSameTeam(UTO,Instigator)) {
            if (UTO.LinkHealMult > 0) {
                if(LinkedTo != Who)
                {
                    Unlink();
                }
                LinkedTo = UTO;
                LinkedComponent = HitComponent;
            } else {
                UTO.FailedLinkHeal(Instigator.Controller);
                UnLink();
            }
        } else {
            UTP = UTPawn(Who);
            
            if (UTP != None && AttemptLinkToPawn(UTP, HitComponent)) {
                // do nothing
            } else if (Who != None && !Who.bStatic) {
                UnLink();
            }
        }
    }

    if (LinkedTo != None) {
        // Determine if the link has been broken for another reason
        UTP = UTPawn(LinkedTo);
        if (LinkedTo.bDeleteMe || 
            (UTP != None && (UTP.Health <= 0 || Torlan_LinkGun(UTP.Weapon) == None))) 
        {
            UnLink();
            return;
        }

        // if we were passed in LinkedTo, we know we hit it straight on already, so skip the rest
        if (LinkedTo != Who) {
            StartTrace = Instigator.GetWeaponStartTraceLocation();
            EndTrace = GetLinkedtoLocation();

            // First, check to see if we have skewed too much, or if the LinkedTo pawn has died and
            // we didn't get cleaned up.
            V = Normal(EndTrace - StartTrace);
            if (
                V dot vector(Instigator.GetViewRotation()) < LinkFlexibility || 
                VSize(EndTrace - StartTrace) > 1.5 * WeaponRange 
            )
            {
                UnLink();
                return;
            }

            //  If something is blocking us and the actor, drop the link
            HitActor = Trace(HitLocation, HitNormal, EndTrace, StartTrace, true);
            if (HitActor != none && HitActor != LinkedTo) {
                UnLink(true);        // In this case, use a delayed UnLink
            }
        }
    }

    // if we are linked, make sure the proper flash location is set
    if (LinkedTo != None) {
        SetFlashLocation(GetLinkedtoLocation());
    }
}

function ConsumeAmmo(byte FireModeNum) {
    LastWeaponFireTime = WorldInfo.TimeSeconds;
    
    ConsumeBeamAmmo(float(ShotCost[FireModeNum]) * GetPlasmaDamageBonus());
}


function ConsumeBeamAmmo(float Amount) {
    local UTWeap_LinkGun gun;
    local array<UTWeap_LinkGun> LinkedWeapons;
    
    GetLinkedWeapons(LinkedWeapons);
    
    Amount /= (LinkStrength + 1);
    Amount *= FMax(1.1 - 0.1*LinkStrength, 0.6);
    
    // self consumes ammo twice, it's intentional
    LinkedConsumeBeamAmmo(Amount);
    foreach LinkedWeapons(gun) {
        Torlan_LinkGun(gun).LinkedConsumeBeamAmmo(Amount);
    }
}

function LinkedConsumeBeamAmmo(float Amount) {
    super.ConsumeBeamAmmo(Amount);
}

function CalcLinkStrength() {
    local array<UTWeap_LinkGun> LinkedWeapons;
    
    local UTWeap_LinkGun LinkGun;
    
    local float LastFirerateMult;
    
    if (WorldInfo.TimeSeconds < LastStrengthCalc + MinTimeBetweenCalcs) {
        return;
    }
    LastStrengthCalc = WorldInfo.TimeSeconds;
    
    GetLinkedWeapons(LinkedWeapons);
    LinkStrength = LinkedWeapons.length;
    
    LastFirerateMult = PawnFirerateMult;
    
    PawnDamageMult = 0.0;
    PawnFirerateMult = 0.0;
    
    if (UTPawn(LinkedTo) == None) {
        foreach LinkedWeapons(LinkGun) {
            PawnDamageMult += LinkGun.Instigator.GetDamageScaling() - 1.0;
            if (UTPawn(LinkGun.Instigator) != None) {
                // note 1/x
                // our FireRate multiplier is different from UTPawn's
                PawnFirerateMult += 1.0/UTPawn(LinkGun.Instigator).FireRateMultiplier - 1.0;
            }
        }
        PawnDamageMult /= LinkStrength;
        PawnFirerateMult /= LinkStrength;
    }

    PawnDamageMult += 1.0;
    PawnFirerateMult += 1.0;
    
    if (LastFirerateMult != PawnFirerateMult) {
        if (UTPawn(Instigator) != None) {
            UTPawn(Instigator).FireRateChanged();
        }
    }
    
    CalcShaftLength();
}

function CalcLinkStrengthForce()
{
    LastStrengthCalc = default.LastStrengthCalc;
    CalcLinkStrength();
}

function PowerUpProbablyChanged()
{
    local Torlan_LinkGun LG;
    for (LG = self; LG.GetNextGun() != None; LG = LG.GetNextGun()) {
        // pass
    }
    LG.CalcLinkStrengthForce();
}

simulated function SetWeaponOverlayFlags(UTPawn OwnerPawn)
{
    super.SetWeaponOverlayFlags(OwnerPawn);

    PowerUpProbablyChanged();
}

function CalcShaftLength() {
    WeaponRange = default.WeaponRange;
    
    switch (LinkStrength) {
    case 1:
        break;
    case 2:
        WeaponRange += 300;
        break;
    case 3:
        WeaponRange += 400;
        break;
    default: // 4, 5, ...
        WeaponRange += 500;
        break;
    }
}

function float GetShaftDamageBonus() {
    if (LinkStrength == 1) {
        return 1.0;
    } else {
        return FMin(1.1 + 0.1 * LinkStrength, 2.0);
    }
}

function float GetPlasmaDamageBonus() {
    return 1.0;
}

function float GetHealBonus() {
    if (LinkStrength == 1) {
        return 1.0;
    } else {
        return FMin(1.05 + 0.05 * LinkStrength, 1.5);
    }
}

simulated function float GetFireInterval(byte FireModeNum)
{
    return FireInterval[FireModeNum] / PawnFirerateMult;
}

function class<Projectile> GetProjectileClass()
{
    return (LinkStrength > 1) ? class'T_LinkPowerPlasma' : Super.GetProjectileClass();
}

simulated function CustomFire() {
    local UTProjectile Proj;
    
    // XXX are we sure we want to recalculate it every time?
    CalcLinkStrength();

    Proj = UTProjectile(ProjectileFire());
    
    if (Role == ROLE_Authority) {
        Proj.Damage *= LinkStrength * PawnDamageMult;
        
        if (ClassIsChildOf(Proj.class, class'T_LinkPowerPlasma')) {
            Proj.Damage *= GetPlasmaDamageBonus();
        }
    }
}

// copy-paste from Epic's code
simulated function ProcessBeamHit(vector StartTrace, vector AimDir, out ImpactInfo TestImpact, float DeltaTime)
{
    local float DamageAmount;
    local vector PushForce, ShotDir, SideDir; //, HitLocation, HitNormal, AttachDir;
    local float HealMult;

    Victim = TestImpact.HitActor;

    // If we are on the server, attempt to setup the link
    if (Role == ROLE_Authority)
    {
        // Try linking
        AttemptLinkTo(Victim, TestImpact.HitInfo.HitComponent);

        // set the correct firemode on the pawn, since it will change when linked
        SetCurrentFireMode(CurrentFireMode);

        // if we do not have a link, set the flash location to whatever we hit
        // (if we do have one, AttemptLinkTo() will set the correct flash location for the Actor we're linked to)
        if (LinkedTo == None)
        {
            SetFlashLocation(TestImpact.HitLocation);
        }

        // cause damage or add health/power/etc.
        bBeamHit = false;

        // compute damage amount
        // XXX are we sure we want to recalculate it every time?
        CalcLinkStrength(); 

        // !!!
        DamageAmount = InstantHitDamage[1] * LinkStrength;
        
        if (LinkedTo != None) {
            DamageAmount *= PawnFirerateMult * PawnDamageMult * GetHealBonus();
            SavedDamage += DamageAmount * DeltaTime;
            DamageAmount = int(SavedDamage);
            SavedAmmoUse += BeamAmmoUsePerSecond * DeltaTime;

            if (DamageAmount >= MinimumDamage) {
                if (!LinkedTo.IsA('UTPawn')) {
                    if (LinkedTo.IsA('UTVehicle') || LinkedTo.IsA('UTGameObjective'))  {
                        // BEGIN HACK: compensate int rounding
                        
                        if (UTVehicle(LinkedTo) != None) {
                            HealMult = UTVehicle(LinkedTo).LinkHealMult;
                        } else if (UTOnslaughtObjective(LinkedTo) != None) {
                            HealMult = UTOnslaughtObjective(LinkedTo).LinkHealMult;
                        } else {
                            HealMult = 1.0;
                        }
                        
                        SavedDamage -= DamageAmount - (DamageAmount*HealMult - int(DamageAmount*HealMult))/HealMult;
                        
                        // END HACK
                        
                        if (LinkedTo.HealDamage(DamageAmount, Instigator.Controller, InstantHitDamageTypes[1]))
                            ConsumeBeamAmmo(SavedAmmoUse);
                    } else {
                        // otherwise always use ammo
                        // shouldn't happen though
                        SavedDamage -= DamageAmount; 
                        ConsumeBeamAmmo(SavedAmmoUse);
                    }
                } else {
                    SavedDamage -= DamageAmount;
                }
                SavedAmmoUse = 0.0;
            } 
        }
        else 
        {
            // If not on the same team, hurt them
            DamageAmount *= PawnFirerateMult * PawnDamageMult * GetShaftDamageBonus();
            SavedDamage += DamageAmount * DeltaTime;
            DamageAmount = int(SavedDamage);
            SavedAmmoUse += BeamAmmoUsePerSecond * DeltaTime;

            if (DamageAmount >= MinimumDamage)
            { 
                SavedDamage -= DamageAmount; 
                ConsumeBeamAmmo(SavedAmmoUse);
                SavedAmmoUse = 0.0;
                if (Victim != None && !WorldInfo.Game.GameReplicationInfo.OnSameTeam(Victim, Instigator))
                {
                    bBeamHit = !Victim.bWorldGeometry;
                    if (DamageAmount > 0 )
                    {
                        ShotDir = Normal(TestImpact.HitLocation - Location);
                        SideDir = Normal(ShotDir Cross vect(0,0,1));
                        PushForce =  vect(0,0,1) + Normal(SideDir * (SideDir dot (TestImpact.HitLocation - Victim.Location)));
                        PushForce *= (Victim.Physics == PHYS_Walking) ? 0.1*MomentumTransfer : DeltaTime*MomentumTransfer;
                        // we divide by GetDamageScaling because it's handled in PawnDamageMult
                        // and DamageAmount is modified somewhere in Game.ReduceDamage
                        DamageAmount /= Instigator.GetDamageScaling();
                        Victim.TakeDamage(DamageAmount, Instigator.Controller, TestImpact.HitLocation, PushForce, InstantHitDamageTypes[1], TestImpact.HitInfo, self);
                    }
                }
            }

        }
    }
    else // end if (Role == ROLE_Authority)
    {
        // if we do not have a link, set the flash location to whatever we hit
        // (otherwise beam update will override with link location)
        if (LinkedTo == None)
        {
            SetFlashLocation(TestImpact.HitLocation);
        }
        else if (TestImpact.HitActor == LinkedTo && TestImpact.HitInfo.HitComponent != None)
        {
            // the linked component can't be replicated to the client, so set it here
            LinkedComponent = TestImpact.HitInfo.HitComponent;
        }
    }
}


simulated function ActiveRenderOverlays(HUD H) {
    local UTHUD UTH;
    local vector2d Pos;

    super.ActiveRenderOverlays(H);
    
    if (LinkStrength > LastLinkStrength) {
        LinkPulseTime = WorldInfo.TimeSeconds;
    }
    LastLinkStrength = LinkStrength;

    UTH = UTHUD(H);
    if (UTH != None) {    
        if (LinkStrength > 1 || true) {
            /* Where:
            0.8 - scaling in terms of ammo bar
            7 - vertical offset
            */
            
            Pos = UTH.ResolveHudPosition(UTH.AmmoPosition, UTH.AmmoBGCoords.UL*0.8, UTH.AmmoBGCoords.VL);
            
            H.Canvas.SetPos(Pos.X, Pos.Y - ((UTH.AmmoBarOffsetY + UTH.AmmoBGCoords.VL - 7)  * UTH.ResolutionScale));
            H.Canvas.DrawColorizedTile(
                UTH.AltHudTexture,
                UTH.AmmoBGCoords.UL * UTH.ResolutionScale * 0.8,
                UTH.AmmoBGCoords.VL * UTH.ResolutionScale,
                UTH.AmmoBGCoords.U,
                UTH.AmmoBGCoords.V,
                UTH.AmmoBGCoords.UL,
                UTH.AmmoBGCoords.VL,
                MakeLinearColor(0, 1, 0, 1)
            );
            
            H.Canvas.DrawColor = H.WhiteColor;

            UTH.DrawGlowText(
                (LinkStrength>1?"+":"")$(LinkStrength - 1),
                Pos.X + (UTH.AmmoTextOffsetX * 0.8 * UTH.ResolutionScale),
                Pos.Y - ((UTH.AmmoBarOffsetY + UTH.AmmoTextOffsetY + UTH.AmmoBGCoords.VL - 7) * UTH.ResolutionScale),
                58 * UTH.ResolutionScale,
                LinkPulseTime,
                true
            );
        }
    }
}

simulated function Activate()
{
    super.Activate();

    if (Pawn(Owner) != None && 
        UTPlayerController(Pawn(Owner).Controller) != None &&
        UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD) != None)
    {
        UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD).HeroMeterOffsetY = 
            UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD).default.HeroMeterOffsetY * 2;
    }
}

auto state Inactive {
    simulated event BeginState(name PreviousStateName) 
    {
        super.BeginState(PreviousStateName);

        if (Pawn(Owner) != None && 
            UTPlayerController(Pawn(Owner).Controller) != None &&
            UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD) != None)
        {
            UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD).HeroMeterOffsetY = 
                UTHUD(UTPlayerController(Pawn(Owner).Controller).myHUD).default.HeroMeterOffsetY;
        }
    }
}


simulated function GetWeaponDebug(out array<string> DebugInfo) {
    super.GetWeaponDebug(DebugInfo);
    
    DebugInfo[DebugInfo.Length] = "Link Strength:" @ LinkStrength;
    DebugInfo[DebugInfo.Length] = "LinkedTo:" @ LinkedTo;
    DebugInfo[DebugInfo.Length] = "LinkedComponent:" @ LinkedComponent;
    DebugInfo[DebugInfo.Length] = "PawnDamageMult:" @ PawnDamageMult;
    DebugInfo[DebugInfo.Length] = "PawnFirerateMult:" @ PawnFirerateMult;
}



defaultproperties
{
    AttachmentClass=class'Torlan_LinkGunAttachment'
    
    WeaponFireTypes(0)=EWFT_Custom
    
    LinkFlexibility=0.62
    
    MinTimeBetweenCalcs=0.5
    LastStrengthCalc=-10.0 // should be at lest less than MinTime
    
    // defaults
    PawnDamageMult=1.0
    PawnFirerateMult=1.0
    LastLinkStrength=1.0
    
    WeaponProjectiles(0)=class'T_LinkPlasma'
}
