class Torlan_LinkGunAttachment extends UTAttachment_LinkGun;

var Texture2D LinkIcon;


simulated function PostRenderFor(PlayerController PC, Canvas Canvas, Object.Vector CameraPosition, Object.Vector CameraDir) {
    local UTPawn Attachment_Pawn;//, PC_Pawn;
    local UTHUD H;
    
    local Vector screenLoc;
    local float SizeX, SizeY;
    
    local float YL; // as in DrawPlayerBeacon
    
    local UTWeapon Weap;
    
    super.PostRenderFor(PC, Canvas, CameraPosition, CameraDir);
    
    Attachment_Pawn = UTPawn(Owner);
    
    H = UTHUD(PC.myHUD);

    
    if (
        WorldInfo.GRI.OnSameTeam(Attachment_Pawn, PC) &&
        Attachment_Pawn.DrivenVehicle == None &&
        (WorldInfo.TimeSeconds - Attachment_Pawn.LastRenderTime < 0.2) &&
        Attachment_Pawn.bPostRenderTraceSucceeded &&
        ((vector(H.PlayerOwner.Rotation) dot (Attachment_Pawn.Location - H.PlayerOwner.ViewTarget.Location)) > 0.f)
        )
    {
        screenLoc = Canvas.Project(Attachment_Pawn.Location + Attachment_Pawn.GetCollisionHeight()*vect(0,0,1));
        
        
        if (screenLoc.X < 0 ||
            screenLoc.X >= Canvas.ClipX ||
            screenLoc.Y < 0 ||
            screenLoc.Y >= Canvas.ClipY
            )
        {
            return;
        }
        
        
        if (UTPawn(PC.Pawn) != None) {
            Weap = UTWeapon(UTPawn(PC.Pawn).Weapon);
            if ((Weap != None) && Weap.CoversScreenSpace(screenLoc, Canvas)) {
                return;
            }
        }
        
        
        if ((UTVehicle_Hoverboard(PC.Pawn) != None) && UTVehicle_Hoverboard(PC.Pawn).CoversScreenSpace(screenLoc, Canvas)) {
            return;
        }    

        
        // now we essentially repeat UTHUD:DrawPlayerBeacon
        
        // ignore SizeX
        Canvas.Font = H.HudFonts[0];
        Canvas.StrLen(Attachment_Pawn.PlayerReplicationInfo.GetPlayerAlias(), SizeX, YL);
        
        if (Attachment_Pawn.PlayerReplicationInfo == H.CharPRI) {
            YL += (34 * Canvas.ClipX/1280);
        }
        
        
        // XXX assuming that LinkIcon is square
        SizeX = 12;
        SizeY = SizeX;
        
        Canvas.SetPos(screenLoc.X - SizeX/2, screenLoc.Y - SizeY - 1.2*YL);
        Canvas.SetDrawColor(0x00, 0xE0, 0x00);
        Canvas.DrawTile(LinkIcon, SizeX, SizeY, 0, 0, LinkIcon.SizeX, LinkIcon.SizeY);
    }
}


simulated function AttachTo(UTPawn OwnerPawn) {
    local PlayerController PC;
    
    
    super.AttachTo(OwnerPawn);
    
    foreach LocalPlayerControllers(class'PlayerController', PC) {
        if (UTHUD(PC.MyHUD) != None && Owner != PC.Pawn) {
            UTHUD(PC.MyHUD).AddPostRenderedActor(self);
        }
    }
}

simulated function DetachFrom(SkeletalMeshComponent MeshCpnt) {
    local PlayerController PC;

    
    super.DetachFrom(MeshCpnt);
    
    foreach LocalPlayerControllers(class'PlayerController', PC) {
        if (UTHUD(PC.MyHUD) != None && Owner != PC.Pawn) {
            UTHUD(PC.MyHUD).RemovePostRenderedActor(self);
        }
    }
}

simulated function PostBeginPlay() {
    super.PostBeginPlay();
    
    //if (WorldInfo.NetMode != NM_DedicatedServer) {
    //    LinkIcon = Texture2D(DynamicLoadObject(class.GetPackageName()$".TorlanRu_Content_01.LinkBeaconT", class'Texture2D'));
    //}
}

// copy-paste from Epic's code
simulated function SetImpactedActor(Actor HitActor, vector HitLocation, vector HitNormal) {
    local SoundCue DesiredLinkHitSound;
    local MaterialInstanceTimeVarying MITV_Decal;

    if (WorldInfo.NetMode != NM_DedicatedServer) {
        if (HitActor != None) {
            DesiredLinkHitSound = (UTPawn(HitActor) != None && !WorldInfo.GRI.OnSameTeam(HitActor, Instigator))
                            ? SoundCue'A_Weapon_Link.Cue.A_Weapon_Link_AltFireImpactFleshCue'
                            : SoundCue'A_Weapon_Link.Cue.A_Weapon_Link_AltFireImpactCue';
            if(!bHittingWall) {
                bHittingWall = UTPawn(HitActor) == none;
                if(bHittingwall) {
                    HitWallEffect = Spawn(class'UTEmitter', self,, HitLocation);
                    HitWallEffect.SetTemplate(WallHitTemplate);
                }
            }

            if (LinkHitSound == None || LinkHitSound.SoundCue != DesiredLinkHitSound) {
                if (LinkHitSound != None) {
                    LinkHitSound.FadeOut(0.1f, 0.0f);
                }
                if (Instigator != None) {
                    LinkHitSound = Instigator.CreateAudioComponent(DesiredLinkHitSound, false, true);
                }
                if (LinkHitSound != None) {
                    LinkHitSound.FadeIn(0.1f, 1.0f);
                }
            }

            if (BeamEndpointEffect != None) {
                BeamEndpointEffect.SetRotation(rotator(HitNormal));
            }
        } else if (LinkHitSound != None) {
            LinkHitSound.FadeOut(0.1f,0.0f);
            LinkHitSound = None;
            bHittingWall = false;
        }
        if (bHittingWall) {
            if (HitWallEffect != none) {
                HitWallEffect.SetRotation(rotator(HitNormal));
                HitWallEffect.SetLocation(HitLocation);
            }

            // Apply beam decal
            if ((TicksBetweenDecals <= TicksSinceLastDecal) && (Terrain(HitActor) == None) && (UTOnslaughtNodeObjective(HitActor) == None) ) {
                if (MaterialInstanceTimeVarying(BeamDecal) != none ) {
                    MITV_Decal = new(self) class'MaterialInstanceTimeVarying';
                    MITV_Decal.SetParent(BeamDecal);
                    WorldInfo.MyDecalManager.SpawnDecal(MITV_Decal, HitLocation, rotator(-HitNormal), DecalWidth, DecalHeight, 10.0, false);
                    MITV_Decal.SetScalarStartTime(DecalDissolveParamName, DurationOfDecal);
                } else {
                    WorldInfo.MyDecalManager.SpawnDecal(BeamDecal, HitLocation, rotator(-HitNormal), DecalWidth, DecalHeight, 10.0, true);
                }
                TicksSinceLastDecal = 0;
            } else {
                ++TicksSinceLastDecal;
            }
        }
        else {
            HitWallEffect = none;
        }
    }
}

defaultproperties
{
    PoweredUpEffect=None
    bPostRenderIfNotVisible=true

    LinkIcon=Texture2D'TorlanRu_Content_02.LinkBeaconT'
}
